"""GCMD DIF 10.2 for identifying updates in metadata over time."""

import copy
import json
import sys
from collections import OrderedDict
from logging import getLogger

from xmltodict import unparse

from envidat.api.v1 import get_protocol_and_domain

log = getLogger(__name__)


def convert_dif(metadata_record: dict) -> str:
    """Generate GCMD DIF 10.2 formatted XML string.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.

    Returns:
        str: XML formatted string compatible with GCMD DIF 10.2 standard

    """
    try:
        converted_dict = dif_convert_dataset(
            metadata_record
        )  # Convert package to OrderedDict
        return unparse(converted_dict, pretty=True)  # Convert OrderedDict to XML
    except ValueError as e:
        log.error(e)
        log.error("Cannot convert package to GCMD DIF 10.2 format.")
        raise ValueError("Failed to convert package to GCMD DIF 10.2 format.")


def dif_convert_dataset(dataset_dict: dict):
    """Create the DIF string from API dictionary."""
    # some values only as custom fields
    extras_dict = extras_as_dict(dataset_dict.get("extras", {}))
    dif_extras = ["science_keywords", "purpose"]

    dif_metadata_dict = OrderedDict()

    # Header
    dif_metadata_dict["@xmlns"] = "http://gcmd.gsfc.nasa.gov/Aboutus/xml/dif/"
    dif_metadata_dict["@xmlns:dif"] = "http://gcmd.gsfc.nasa.gov/Aboutus/xml/dif/"
    dif_metadata_dict["@xmlns:xsi"] = "http://www.w3.org/2001/XMLSchema-instance"
    namespace = "http://gcmd.gsfc.nasa.gov/Aboutus/xml/dif/"
    schema = "http://gcmd.gsfc.nasa.gov/Aboutus/xml/dif/dif_v10.2.xsd"
    dif_metadata_dict["@xsi:schemaLocation"] = f"{namespace} {schema}"

    # Entry_ID
    dif_metadata_dict["Entry_ID"] = OrderedDict()
    dif_metadata_dict["Entry_ID"]["Short_Name"] = dataset_dict.get("name", "")
    dif_metadata_dict["Entry_ID"]["Version"] = dataset_dict.get("version", "1.0")

    # Version_Description

    # Entry_Title
    dif_metadata_dict["Entry_Title"] = dataset_dict.get("title", "")

    # Dataset_Citation
    dif_metadata_dict["Dataset_Citation"] = OrderedDict()

    # "Dataset_Creator" organization
    author_names = []
    try:
        for author in json.loads(dataset_dict.get("author", "[]")):
            author_name = ""
            if author.get("given_name"):
                author_name += author["given_name"].strip() + " "
            author_names += [author_name + author["name"].strip()]
    except ValueError:
        pass

    if author_names:
        dif_metadata_dict["Dataset_Citation"]["Dataset_Creator"] = ", ".join(
            author_names
        )

    # "Dataset_Editor" maintainer
    try:
        maintainer = json.loads(dataset_dict.get("maintainer", "{}"))
        maintainer_name = ""
        if maintainer.get("given_name"):
            maintainer_name += maintainer["given_name"].strip() + " "
        maintainer_name += maintainer["name"]
        dif_metadata_dict["Dataset_Citation"][
            "Dataset_Editor"
        ] = maintainer_name.strip()
    except ValueError:
        pass

    # "Dataset_Title"
    dif_metadata_dict["Dataset_Citation"]["Dataset_Title"] = dataset_dict.get(
        "title", ""
    )

    # "Dataset_Series_Name"
    # "Dataset_Release_Date"
    publication_year = json.loads(dataset_dict.get("publication", "{}")).get(
        "publication_year", ""
    )
    dif_metadata_dict["Dataset_Citation"]["Dataset_Release_Date"] = publication_year

    # "Dataset_Release_Place"
    dif_metadata_dict["Dataset_Citation"][
        "Dataset_Release_Place"
    ] = "Birmensdorf, Switzerland"

    # "Dataset_Publisher"
    dif_metadata_dict["Dataset_Citation"]["Dataset_Publisher"] = (
        json.loads(dataset_dict.get("publication", "{}")).get("publisher", "").strip()
    )

    # "Version"
    dif_metadata_dict["Dataset_Citation"]["Version"] = dataset_dict.get("version", "")

    # "Issue_Identification"
    # "Data_Presentation_Form"
    dif_metadata_dict["Dataset_Citation"]["Data_Presentation_Form"] = ",".join(
        get_resource_formats(dataset_dict)
    )

    # "Other_Citation_Details"
    # "Persistent_Identifier"
    doi = dataset_dict.get("doi", "")
    if doi:
        identifier = OrderedDict()
        identifier["Type"] = "DOI"
        identifier["Identifier"] = "doi:" + doi.strip()
        dif_metadata_dict["Dataset_Citation"]["Persistent_Identifier"] = identifier

    # "Online_Resource"
    protocol, host = get_protocol_and_domain()
    package_name = dataset_dict.get("name", "")
    package_url = f"{protocol}://{host}/dataset/{package_name}"
    dif_metadata_dict["Dataset_Citation"]["Online_Resource"] = package_url

    # "Personnel"
    maintainer = json.loads(dataset_dict.get("maintainer", "{}"))
    dif_metadata_dict["Personnel"] = OrderedDict()
    dif_metadata_dict["Personnel"]["Role"] = "TECHNICAL CONTACT"
    dif_metadata_dict["Personnel"]["Contact_Person"] = OrderedDict()
    dif_metadata_dict["Personnel"]["Contact_Person"]["First_Name"] = maintainer.get(
        "given_name", maintainer.get("name", "").strip().split(" ")[0]
    ).strip()
    dif_metadata_dict["Personnel"]["Contact_Person"]["Last_Name"] = (
        maintainer.get("name", "").strip().split(" ")[-1]
    )
    dif_metadata_dict["Personnel"]["Contact_Person"]["Email"] = maintainer.get(
        "email", ""
    ).strip()

    # Science_Keywords (M)*
    science_keywords = get_science_keywords(dataset_dict, extras_dict)
    dif_metadata_dict["Science_Keywords"] = OrderedDict()
    dif_metadata_dict["Science_Keywords"]["Category"] = science_keywords[0]
    dif_metadata_dict["Science_Keywords"]["Topic"] = science_keywords[1]
    dif_metadata_dict["Science_Keywords"]["Term"] = science_keywords[2]
    if len(science_keywords) > 3:
        dif_metadata_dict["Science_Keywords"]["Variable_Level_1"] = science_keywords[3]
    if len(science_keywords) > 4:
        dif_metadata_dict["Science_Keywords"]["Variable_Level_2"] = science_keywords[4]

    # "ISOTopicCategoryType"
    # select from https://gcmd.nasa.gov/add/difguide/iso_topic_category.html
    dif_metadata_dict["ISO_Topic_Category"] = "environment"

    # Ancillary_Keyword
    dif_metadata_dict["Ancillary_Keyword"] = get_keywords(dataset_dict)

    # "Platform"
    dif_metadata_dict["Platform"] = OrderedDict()
    dif_metadata_dict["Platform"]["Type"] = "Not provided"
    dif_metadata_dict["Platform"]["Short_Name"] = "Not provided"
    dif_metadata_dict["Platform"]["Instrument"] = {"Short_Name": "Not provided"}

    # Temporal_Coverage
    dif_metadata_dict["Temporal_Coverage"] = OrderedDict()

    # default set to publication year
    dif_metadata_dict["Temporal_Coverage"]["Single_DateTime"] = (
        publication_year + "-01-01"
    )

    # "Dataset_Progress" draft or private -> IN WORK, doi -> COMPLETE (otherwise empty)
    if dataset_dict.get("private", False) or dataset_dict.get("num_resources", 0) == 0:
        dif_metadata_dict["Dataset_Progress"] = "IN WORK"
    elif dataset_dict.get("doi", ""):
        dif_metadata_dict["Dataset_Progress"] = "COMPLETE"

    # Spatial_Coverage
    dif_metadata_dict["Spatial_Coverage"] = OrderedDict()
    # "Spatial_Coverage_Type"
    # "Granule_Spatial_Representation"
    dif_metadata_dict["Spatial_Coverage"][
        "Granule_Spatial_Representation"
    ] = "CARTESIAN"

    # <xs:element name="Zone_Identifier" type="xs:string" minOccurs="0"/>

    # "Geometry" [1]
    try:
        spatial = json.loads(dataset_dict.get("spatial", "{}"))
    except ValueError:
        spatial = {}
    if spatial:
        dif_metadata_dict["Spatial_Coverage"]["Geometry"] = OrderedDict()
        dif_metadata_dict["Spatial_Coverage"]["Geometry"][
            "Coordinate_System"
        ] = "CARTESIAN"

        coordinates = spatial.get("coordinates", [])
        if coordinates:
            bounding_rectangle = get_bounding_rectangle_dict(spatial)
            dif_metadata_dict["Spatial_Coverage"]["Geometry"][
                "Bounding_Rectangle"
            ] = bounding_rectangle

        # <xs:element name="Point" type="Point"/>
        if spatial.get("type") == "Point":
            bound_box_coordinates = get_bounding_rectangle(
                spatial.get("coordinates", [])
            )
            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Point"] = OrderedDict()
            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Point"][
                "Point_Longitude"
            ] = bound_box_coordinates[0]
            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Point"][
                "Point_Latitude"
            ] = bound_box_coordinates[3]
            # latitude = bound_box_coordinates[3]
        elif spatial.get("type") == "MultiPoint":
            points = []
            for coordinate_pair in spatial.get("coordinates", []):
                point = OrderedDict()
                point["Point_Longitude"] = str(coordinate_pair[0])
                point["Point_Latitude"] = str(coordinate_pair[1])
                points += [point]
            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Point"] = points
        elif spatial.get("type") == "Polygon":
            # <xs:element name="Polygon" type="GPolygon"/>
            points = []
            for coordinate_pair in spatial.get("coordinates", [])[0]:
                point = OrderedDict()
                point["Point_Longitude"] = str(coordinate_pair[0])
                point["Point_Latitude"] = str(coordinate_pair[1])
                points += [point]
            if len(points) > 1:
                points.pop()

            if is_counter_clockwise(points):
                log.debug(
                    dataset_dict.get("name", "") + " ** Counterclockwise REVERSING!! **"
                )
                points.reverse()
            else:
                log.debug(dataset_dict.get("name", "") + " Clockwise OK")

            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Polygon"] = OrderedDict()
            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Polygon"]["Boundary"] = {
                "Point": points
            }
            dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Polygon"][
                "Center_Point"
            ] = copy.deepcopy(
                dif_metadata_dict["Spatial_Coverage"]["Geometry"]["Bounding_Rectangle"][
                    "Center_Point"
                ]
            )

    # <xs:element name="Orbit_Parameters" type="OrbitParameters" minOccurs="0"/>
    # <xs:element name="Vertical_Spatial_Info" type="VerticalSpatialInfo" minOccurs="0"
    #   maxOccurs="unbounded"/>
    # <xs:element name="Spatial_Info" type="SpatialInfo" minOccurs="0"/>

    # <xs:element name="Location" type="LocationType" minOccurs="0"
    #   maxOccurs="unbounded"/>
    # Cannot know type, could be set to CONTINENT type and then Europe (?)
    # <xs:element name="Data_Resolution" type="DataResolutionType" minOccurs="0"
    #   maxOccurs="unbounded"/>

    # Project
    dif_metadata_dict["Project"] = {"Short_Name": "Not provided"}

    # <xs:element name="Quality" type="QualityType" minOccurs="0"/>
    dif_metadata_dict["Access_Constraints"] = "Public access to the data"

    dataset_restrictions = get_resource_restrictions(dataset_dict)
    if "registered" in dataset_restrictions:
        dif_metadata_dict[
            "Access_Constraints"
        ] = "Registration is required to access the data"
    elif (
        ("any_organization" in dataset_restrictions)
        or ("same_organization" in dataset_restrictions)
        or ("only_allowed_users" in dataset_restrictions)
    ):
        dif_metadata_dict["Access_Constraints"] = "Access to the data upon request"

    # "Use_Constraints"
    license = dataset_dict.get("license_title", None)
    if not license:
        license = "Open Data Commons Open Database License (ODbL)"
    license_url = dataset_dict.get(
        "license_url", "http://www.opendefinition.org/licenses/odc-odbl"
    )
    dif_metadata_dict["Use_Constraints"] = (
        'Usage constraints defined by the license "'
        + license.strip()
        + '", see '
        + license_url
    )

    # Dataset_Language
    dif_metadata_dict["Dataset_Language"] = get_dif_language_code(
        dataset_dict.get("language", "en")
    )

    # "Originating_Center"
    dif_metadata_dict["Originating_Center"] = dataset_dict.get("organization", {}).get(
        "title", ""
    )

    # Organization
    dif_metadata_dict["Organization"] = OrderedDict()

    # "Organization_Type" * DISTRIBUTOR/ARCHIVER/ORIGINATOR/PROCESSOR
    dif_metadata_dict["Organization"]["Organization_Type"] = "DISTRIBUTOR"

    # "Organization_Name" "Short_Name" "Long_Name"
    dif_metadata_dict["Organization"]["Organization_Name"] = OrderedDict()
    dif_metadata_dict["Organization"]["Organization_Name"]["Short_Name"] = "WSL"
    dif_metadata_dict["Organization"]["Organization_Name"][
        "Long_Name"
    ] = "Swiss Federal Institute for Forest, Snow and Landscape Research WSL"

    # <xs:element name="Hours_Of_Service" type="xs:string" minOccurs="0"/>
    # <xs:element name="Instructions" type="xs:string" minOccurs="0"/>
    #  "Organization_URL"
    dif_metadata_dict["Organization"]["Organization_URL"] = "https://www.wsl.ch"

    # <xs:element name="Dataset_ID" type="xs:string" minOccurs="0"
    #   maxOccurs="unbounded"/>
    # <xs:element name="Personnel" type="OrgPersonnelType" maxOccurs="unbounded"/>
    dif_metadata_dict["Organization"]["Personnel"] = OrderedDict()
    dif_metadata_dict["Organization"]["Personnel"]["Role"] = "DATA CENTER CONTACT"
    dif_metadata_dict["Organization"]["Personnel"]["Contact_Group"] = OrderedDict()
    dif_metadata_dict["Organization"]["Personnel"]["Contact_Group"]["Name"] = "EnviDat"
    dif_metadata_dict["Organization"]["Personnel"]["Contact_Group"][
        "Email"
    ] = "envidat@wsl.ch"

    # <xs:element name="Distribution" type="DistributionType" minOccurs="0"
    #   maxOccurs="unbounded"/>
    # <xs:element name="Multimedia_Sample" type="MultimediaSampleType"
    #   minOccurs="0" maxOccurs="unbounded"/>

    # <xs:element name="Reference" type="ReferenceType" minOccurs="0"
    #   maxOccurs="unbounded"/>
    # Find paper citation in the description and parse it to this element

    thumbnail_url = "https://www.envidat.ch/envidat_thumbnail.png"
    dif_metadata_dict["Multimedia_Sample"] = OrderedDict()
    dif_metadata_dict["Multimedia_Sample"]["URL"] = thumbnail_url

    # Summary
    dif_metadata_dict["Summary"] = OrderedDict()
    # Abstract
    dif_metadata_dict["Summary"]["Abstract"] = (
        dataset_dict.get("notes", "").replace("\n", " ").replace("\r", " ").strip()
    )
    # "Purpose"
    # purpose = get_ignore_case(extras_dict, 'purpose')
    # if purpose:
    # dif_metadata_dict['Summary']['Purpose'] = get_or_missing(
    #     extras_dict, 'purpose', ignore_case=True
    # )
    # Related_URL
    dif_metadata_dict["Related_URL"] = {"URL": package_url}

    # <xs:element name="Metadata_Association" type="MetadataAssociationType"
    #   minOccurs="0" maxOccurs="unbounded"/>
    # <xs:element name="IDN_Node" type="IDNNodeType" minOccurs="0"
    #   maxOccurs="unbounded"/>
    # <xs:element name="Originating_Metadata_Node" type="OriginatingMetadataNodeType"
    #   minOccurs="0"/>

    # Metadata_Name
    dif_metadata_dict["Metadata_Name"] = "gcmd_dif"

    # Metadata_Version
    dif_metadata_dict["Metadata_Version"] = "VERSION 10.2"

    # <xs:element name="DIF_Revision_History" type="DIFRevisionHistoryType"
    #   minOccurs="0"/>

    # Metadata_Dates (M)
    dif_metadata_dict["Metadata_Dates"] = OrderedDict()

    metadata_created = dataset_dict.get("metadata_created")
    metadata_modified = dataset_dict.get("metadata_modified")
    dif_metadata_dict["Metadata_Dates"]["Metadata_Creation"] = metadata_created
    dif_metadata_dict["Metadata_Dates"]["Metadata_Last_Revision"] = metadata_modified
    dif_metadata_dict["Metadata_Dates"]["Data_Creation"] = metadata_created
    dif_metadata_dict["Metadata_Dates"]["Data_Last_Revision"] = metadata_modified

    # "Private"
    if dataset_dict.get("private", False):
        dif_metadata_dict["Private"] = "True"
    else:
        dif_metadata_dict["Private"] = "False"

    # "Additional_Attributes"
    # Maybe the authors should go here

    # <xs:element name="Product_Level_Id" type="ProcessingLevelIdType" minOccurs="0"/>
    # <xs:element name="Collection_Data_Type" type="CollectionDataTypeEnum"
    #   minOccurs="0" maxOccurs="unbounded"/>
    # <xs:element name="Product_Flag" type="ProductFlagEnum" minOccurs="0"/>

    # "Extended_Metadata"
    extended_metadata = []
    for key in extras_dict:
        if key.lower() not in dif_extras:
            value = extras_dict[key]
            metadata = OrderedDict()
            metadata["Name"] = key.strip()
            metadata["Type"] = "String"
            metadata["Value"] = value.strip()
            extended_metadata += [metadata]
    if len(extended_metadata) > 0:
        dif_metadata_dict["Extended_Metadata"] = {"Metadata": extended_metadata}

    # Root element
    gcmd_dif_dict = OrderedDict()
    gcmd_dif_dict["DIF"] = dif_metadata_dict

    return gcmd_dif_dict


def get_keywords(data_dict):
    """Extract keywords from tags."""
    keywords = []
    for tag in data_dict.get("tags", []):
        name = tag.get("display_name", "").upper()
        keywords += [name]
    return keywords


def get_science_keywords(data_dict, extras_dict):
    """Guess keywords from organization."""
    default_keywords = [
        "EARTH SCIENCE",
        "CLIMATE INDICATORS",
        "LAND SURFACE/AGRICULTURE INDICATORS",
    ]

    # check if defined in extras, comma-separated
    custom_keywords = (
        get_ignore_case(extras_dict, "science_keywords").upper().split(",")
    )
    if len(custom_keywords) >= 3:
        return custom_keywords

    # map to organization
    dataset_organization = data_dict.get("organization", {}).get("name", "")

    # possible topics: AGRICULTURE, ATMOSPHERE, BIOSPHERE, BIOLOGICAL CLASSIFICATION,
    # CLIMATE INDICATORS, CRYOSPHERE, HUMAN DIMENSIONS, LAND SURFACE, OCEANS,
    # PALEOCLIMATE, SOLID EARTH, SPECTRAL/ENGINEERING, SUN-EARTH INTERACTIONS,
    # TERRESTRIAL HYDROSPHERE
    organizations_keywords_dict = {
        "biodiversity-and-conservation-biology": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOLOGICAL DYNAMICS",
            "COMMUNITY DYNAMICS",
            "BIODIVERSITY FUNCTIONS",
        ],
        "cces": ["EARTH SCIENCE", "CLIMATE INDICATORS", "ATMOSPHERIC/OCEAN INDICATORS"],
        "clench": [
            "EARTH SCIENCE",
            "CLIMATE INDICATORS",
            "ATMOSPHERIC/OCEAN INDICATORS",
        ],
        "community-ecology": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOLOGICAL DYNAMICS",
            "COMMUNITY DYNAMICS",
        ],
        "conservation-biology": [
            "EARTH SCIENCE",
            "AGRICULTURE",
            "FOREST SCIENCE",
            "FOREST CONSERVATION",
        ],
        "cryos": ["EARTH SCIENCE", "CRYOSPHERE", "SNOW/ICE"],
        "d-baug": ["EARTH SCIENCE", "SPECTRAL/ENGINEERING", "PLATFORM CHARACTERISTICS"],
        "usys": ["EARTH SCIENCE", "CLIMATE INDICATORS", "ATMOSPHERIC/OCEAN INDICATORS"],
        "dynamic-macroecology": ["EARTH SCIENCE", "BIOSPHERE", "ECOLOGICAL DYNAMICS"],
        "ecosystems-dynamics": ["EARTH SCIENCE", "BIOSPHERE", "ECOSYSTEMS"],
        "epfl": ["EARTH SCIENCE", "CLIMATE INDICATORS", "ATMOSPHERIC/OCEAN INDICATORS"],
        "ethz": ["EARTH SCIENCE", "CLIMATE INDICATORS", "ATMOSPHERIC/OCEAN INDICATORS"],
        "feh": ["EARTH SCIENCE", "AGRICULTURE", "FOOD SCIENCE"],
        "forema": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOSYSTEMS",
            "TERRESTRIAL ECOSYSTEMS",
            "FOREST",
        ],
        "forest-dynamics": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOSYSTEMS",
            "TERRESTRIAL ECOSYSTEMS",
            "FORESTS",
        ],
        "forest-soils-and-biogeochemistry": ["EARTH SCIENCE", "LAND SURFACE", "SOILS"],
        "gebirgshydrologie": [
            "EARTH SCIENCE",
            "TERRESTRIAL HYDROSPHERE",
            "SURFACE WATER",
        ],
        "gis": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOSYSTEMS",
            "ANTHROPOGENIC/HUMAN INFLUENCED ECOSYSTEMS",
        ],
        "hazri": ["EARTH SCIENCE", "SOLID EARTH", "NATURAL HAZARDS"],
        "ibp": ["EARTH SCIENCE", "CLIMATE INDICATORS", "ATMOSPHERIC/OCEAN INDICATORS"],
        "landscape-dynamics": ["EARTH SCIENCE", "LAND SURFACE", "LANDSCAPE"],
        "lwf": ["EARTH SCIENCE", "CLIMATE INDICATORS", "ATMOSPHERIC/OCEAN INDICATORS"],
        "mountain-ecosystems": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOSYSTEMS",
            "TERRESTRIAL ECOSYSTEMS",
        ],
        "nfi": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOSYSTEMS",
            "TERRESTRIAL ECOSYSTEMS",
            "FORESTS",
        ],
        "plant-animal-interactions": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOLOGICAL DYNAMICS",
            "SPECIES/POPULATION INTERACTIONS",
        ],
        "remote-sensing": [
            "EARTH SCIENCE",
            "CLIMATE INDICATORS",
            "ATMOSPHERIC/OCEAN INDICATORS",
        ],
        "resource-analysis": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "FOREST SCIENCE",
            "FOREST CONSERVATION",
        ],
        "slf": ["EARTH SCIENCE", "CRYOSPHERE", "SNOW/ICE"],
        "stand-dynamics-and-silviculture": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "SILVICULTURE",
        ],
        "swissforestlab-swissfl": [
            "EARTH SCIENCE",
            "BIOSPHERE",
            "ECOSYSTEMS",
            "TERRESTRIAL ECOSYSTEMS",
            "FORESTS",
        ],
        "vaw": ["EARTH SCIENCE", "TERRESTRIAL HYDROSPHERE", "GLACIERS/ICE SHEETS"],
        "wsl": [
            "EARTH SCIENCE",
            "CLIMATE INDICATORS",
            "LAND SURFACE/AGRICULTURE INDICATORS",
        ],
    }

    science_keywords = organizations_keywords_dict.get(
        dataset_organization, default_keywords
    )

    return science_keywords


def get_ignore_case(data_dict, tag, ignore_blanks=True):
    """Get value, case agnostic."""
    tag_lower = tag.lower()
    if ignore_blanks:
        tag_lower = tag_lower.replace(" ", "")
    tag_key = ""
    for key in data_dict.keys():
        key_lower = key.lower()
        if ignore_blanks:
            key_lower = key_lower.replace(" ", "")
        if key_lower == tag_lower:
            tag_key = key
            break
    return data_dict.get(tag_key, "")


def extras_as_dict(extras):
    """Extract API 'extras' field as a simple dictionary."""
    extras_dict = {}
    for extra in extras:
        extras_dict[extra.get("key")] = extra.get("value")
    return extras_dict


def get_resource_formats(dataset_dict):
    """Get resource formats."""
    resource_formats = []
    for resource in dataset_dict.get("resources", []):
        resource_format = resource.get(
            "format", resource.get("mimetype", resource.get("mimetype_inner", ""))
        )
        if resource_format:
            resource_format = resource_format.lower()
            if resource_format not in resource_formats:
                resource_formats += [resource_format]
    return resource_formats


def get_resource_restrictions(dataset_dict):
    """Get resource restrictions."""
    resource_restrictions = []
    for resource in dataset_dict.get("resources", []):
        try:
            restricted = json.loads(resource.get("restricted"))
        except ValueError:
            restricted = {}
        resource_restriction = restricted.get("level", "")
        if resource_restriction:
            resource_restriction = resource_restriction.lower()
            if resource_restriction not in resource_restrictions:
                resource_restrictions += [resource_restriction]
    return resource_restrictions


def get_dif_language_code(code):
    """Translate codes to language full word.

    https://gcmd.nasa.gov/DocumentBuilder/defaultDif10/guide/data_set_language.html

    Options: English; Afrikaans; Arabic; Bosnia; Bulgarian; Chinese; Croation; Czech;
    Danish; Dutch; Estonian; Finnish; French; German; Hebrew; Hungarian; Indonesian;
    Italian; Japanese; Korean; Latvian; Lithuanian; Norwegian; Polish; Portuguese;
    Romanian; Russian; Slovak; Spanish; Ukrainian; Vietnamese
    """
    lang_code = code.lower()[:2]
    lookup_dict = {
        "en": "English",
        "de": "German",
        "it": "Italian",
        "fr": "French",
    }  # , 'ro':'roh'}
    return lookup_dict.get(lang_code, "English").title()


def get_bounding_rectangle(coordinates: list) -> list:
    """Geometry bounding rectangle as coordinate list."""
    flatten_coordinates = coordinates
    while type(flatten_coordinates[0]) is list:
        flatten_coordinates = [
            item for sublist in flatten_coordinates for item in sublist
        ]
    longitude_coords = flatten_coordinates[0:][::2]
    latitude_coords = flatten_coordinates[1:][::2]
    return [
        min(longitude_coords),
        max(longitude_coords),
        min(latitude_coords),
        max(latitude_coords),
    ]


def get_bounding_rectangle_dict(spatial_dict: dict) -> dict:
    """Geometry bounding rectangle as value dictionary."""
    bound_box_coordinates = get_bounding_rectangle(spatial_dict.get("coordinates", []))

    bounding_rectangle = OrderedDict()
    bounding_rectangle["Center_Point"] = OrderedDict()
    bounding_rectangle["Center_Point"]["Point_Longitude"] = str(
        (bound_box_coordinates[1] + bound_box_coordinates[0]) / 2.0
    )
    bounding_rectangle["Center_Point"]["Point_Latitude"] = str(
        (bound_box_coordinates[3] + bound_box_coordinates[2]) / 2.0
    )
    bounding_rectangle["Southernmost_Latitude"] = str(
        max(bound_box_coordinates[2], -90)
    )
    bounding_rectangle["Northernmost_Latitude"] = str(min(bound_box_coordinates[3], 90))
    bounding_rectangle["Westernmost_Longitude"] = str(max(bound_box_coordinates[0], 0))
    bounding_rectangle["Easternmost_Longitude"] = str(
        min(bound_box_coordinates[1], 180)
    )

    return bounding_rectangle


def is_counter_clockwise(points):
    """Check if polygon is counterclockwise / valid."""
    if len(points) < 3:
        return False

    try:
        akku = 0

        for i in range(len(points)):
            p1 = points[i]
            p2 = points[0]

            if i + 1 < len(points):
                p2 = points[i + 1]

            akku += (float(p2["Point_Longitude"]) - float(p1["Point_Longitude"])) * (
                float(p2["Point_Latitude"]) + float(p1["Point_Latitude"])
            )

        if akku >= 0:
            return False
        else:
            return True
    except Exception as e:
        log.error(
            "Unexpected error converting to float (is_counter_clockwise):",
            sys.exc_info()[0],
        )
        log.error(e)

    return False



"""Extra utils used internally and within EnviDat projects."""

import logging
import os
import sys
import urllib.parse
from pathlib import Path
from typing import NoReturn, Union

import requests

log = logging.getLogger(__name__)


def _debugger_is_active() -> bool:
    """Check to see if running in debug mode.

    Returns:
        bool: if a debug trace is present or not.
    """
    gettrace = getattr(sys, "gettrace", lambda: None)
    return gettrace() is not None


def load_dotenv_if_in_debug_mode(env_file: Union[Path, str]) -> NoReturn:
    """Load secret .env variables from repo for debugging.

    Args:
        env_file (Union[Path, str]): String or Path like object pointer to
            secret dot env file to read.
    """
    if not _debugger_is_active():
        return

    try:
        from dotenv import load_dotenv
    except ImportError as e:
        log.error(
            """
            Unable to import dotenv.
            Note: The logger should be invoked after reading the dotenv file
            so that the debug level is by the environment.
            """
        )
        log.error(e)
        raise ImportError from e(
            """
            Unable to import dotenv, is python-dotenv installed?
            Try installing this package using pip install envidat[dotenv].
            """
        )

    secret_env = Path(env_file)
    if not secret_env.is_file():
        log.error(
            """
            Attempted to import dotenv, but the file does not exist.
            Note: The logger should be invoked after reading the dotenv file
            so that the debug level is by the environment.
            """
        )
        raise FileNotFoundError from None(
            f"Attempted to import dotenv, but the file does not exist: {env_file}"
        )
    else:
        load_dotenv(secret_env)


def get_logger() -> logging.basicConfig:
    """Set logger parameters with log level from environment.

    Note:
        Defaults to DEBUG level, unless specified by LOG_LEVEL env var.
    """
    logging.basicConfig(
        level=os.getenv("LOG_LEVEL", default="DEBUG"),
        format=(
            "%(asctime)s.%(msecs)03d [%(levelname)s] "
            "%(name)s | %(funcName)s:%(lineno)d | %(message)s"
        ),
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )
    log.debug("Logger set to STDOUT.")


# TODO use this function to replace get_user_show()
#  in router_publish.send_email_publish_async()
# TODO improve exception handling with more specific exceptions
# TODO write docstring
def get_response_json(
    api_host: str,
    api_path: str,
    query: dict | None = None,
    api_key: str | None = None,
    status_code: int = 200,
) -> dict | None:
    """Get response JSON from EnviDat."""
    load_dotenv_if_in_debug_mode(".env")

    key = None
    # Extract environment variables needed to call API URL
    try:
        host = os.environ(api_host)
        path = os.environ(api_path)
        if api_key:
            key = os.environ(api_key)
    except KeyError as e:
        log.error(f"KeyError: {e} does not exist in environnment vars")
        return None
    except AttributeError as e:
        log.error(e)
        return None

    # Call API and return JSON response
    try:
        # Call API with query params if they are truthy
        if query:
            params = urllib.parse.urlencode(query)
            api_url = f"{host}{path}?{params}"
        else:
            api_url = f"{host}{path}"

        # Add headers to request if key truthy
        if key:
            headers = {"Authorization": key}
            response = requests.get(api_url, headers=headers)
        else:
            response = requests.get(api_url)

        # Handle unexpected response status_code
        # Default expected response status_code is 200
        if response.status_code != status_code:
            log.error(
                f"ERROR call to API returned unexpected response status_code: "
                f"{response.status_code}"
            )
            log.error(f"ERROR message from API: {response.json()}")
            return None

        # Return JSON response
        if response:
            return response.json()

    except ConnectionError as e:
        log.error(e)
        return None

    except Exception as e:
        log.error(e)
        return None


def get_url(url: str) -> requests.Response:
    """Get a URL with additional error handling.

    Args:
        url (str): The URL to GET.
    """
    try:
        log.debug(f"Attempting to get {url}")
        r = requests.get(url)
        r.raise_for_status()
        return r
    except requests.exceptions.ConnectionError as e:
        log.error(f"Could not connect to internet on get: {r.request.url}")
        log.error(e)
    except requests.exceptions.HTTPError as e:
        log.error(f"HTTP response error on get: {r.request.url}")
        log.error(e)
    except requests.exceptions.RequestException as e:
        log.error(f"Request error on get: {r.request.url}")
        log.error(f"Request: {e.request}")
        log.error(f"Response: {e.response}")
    except Exception as e:
        log.error(e)
        log.error(f"Unhandled exception occurred on get: {r.request.url}")

    return None


# TODO refactor this or get_url() as they have similar functionality
def get_url_response(url: str, cookie: str | None = None) -> requests.Response:
    """Get a URL with additional error handling.

    Args:
        url (str): The URL to GET.
        cookie (str | None): Cookie passed to API call in header,
                             default value is None as this argument is not always used
    """
    try:
        log.debug(f"Attempting to get {url}")

        # Call API with cookie in header if it exists
        if cookie:
            headers = {"Cookie": cookie}
            r = requests.get(url, headers=headers)
        # Else call API (without header)
        else:
            r = requests.get(url)

        # Raise HTTP error it if occured
        r.raise_for_status()
    except requests.exceptions.ConnectionError as e:
        log.error(f"Could not connect to internet on get: {r.request.url}")
        log.error(e)
    except requests.exceptions.HTTPError:
        log.error(f"HTTP response error on get: {r.request.url}")
    except requests.exceptions.RequestException as e:
        log.error(f"Request error on get: {r.request.url}")
        log.error(f"Request: {e.request}")
        log.error(f"Response: {e.response}")
    except Exception as e:
        log.error(e)
        log.error(f"Unhandled exception occurred on get: {r.request.url}")

    return r



import logging

# Setup logging
from logging import getLogger

import requests

from envidat.api.v1 import get_envidat_record
from envidat.doi.datacite_publisher import publish_datacite
from envidat.utils import get_response_json

log = getLogger(__name__)
log.setLevel(level=logging.INFO)

# Setup up file log handler
logFileFormatter = logging.Formatter(
    fmt="%(levelname)s %(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)
# TODO implement name of log file as command line argument, assign default name
fileHandler = logging.FileHandler(filename="./logs/datacite_updater.log")
fileHandler.setFormatter(logFileFormatter)
fileHandler.setLevel(level=logging.INFO)
log.addHandler(fileHandler)

# Load config from environment variables
load_dotenv_if_in_debug_mode(".env")


def datacite_update_all_records():
    """Updates existing DOIs for all EnviDat records on DataCite.

    Function converts all EnviDat records to DataCite Metadata Schema 4.4,
    for documentation: https://schema.datacite.org/meta/kernel-4.4/

    For documentation of DataCite API: https://support.datacite.org/docs/api

    For DataCite API reference:
    https://support.datacite.org/reference/introduction

    For documentation of DataCite API endpoint that updates DOIs:
    https://support.datacite.org/reference/put_dois-id
    """
    # Get EnviDat DOIs on DataCite
    dc_dois = get_dc_dois()

    # Handle if dc_dois are None
    if not dc_dois:
        log.error("Failed to get DataCite DOIs")
        return None

    # Get EnviDat record objects for record that are published and have DOIs
    published_records = get_published_record_names_with_dois()

    # Handle if published_records are None
    if not published_records:
        log.error("Failed to get EnviDat published records names that have DOIs")

    # Update or create new DOIs in DataCite for all EnviDat records
    for record in published_records:

        dc_response = datacite_update_one_record(record.get("name"))

        # Log successful update
        # Unsuccessful update error messages logged
        # during call to datacite_update_one_record()
        if dc_response:
            log.info(dc_response)
            continue

    return


def datacite_update_records(record_names: list[str]):
    """Updates existing DOIs for EnviDat records on DataCite.

    ASSUMPTION: Records already exist on DataCite and should be updated.

    Function converts EnviDat records to DataCite Metadata Schema 4.4, for
    documentation: https://schema.datacite.org/meta/kernel-4.4/

    For documentation of DataCite API: https://support.datacite.org/docs/api

    For DataCite API reference:
    https://support.datacite.org/reference/introduction

    For documentation of DataCite API endpoint that updates DOIs:
    https://support.datacite.org/reference/put_dois-id

    Args:
        record_names (list[str]): List of EnviDat records names that should be
            updated. Example: ["mountland-jura", "envidat-lwf-51"]
    """
    # Update DOIs in DataCite for EnviDat record_names
    for name in record_names:

        dc_response = datacite_update_one_record(name)

        # Log successful update
        # Unsuccessful update error messages logged
        # during call to datacite_update_one_record()
        if dc_response:
            log.info(dc_response)
            continue

    return


def datacite_update_one_record(name: str, dc_dois: list[str] = None) -> dict | None:
    """Updates existing DOI for one EnviDat record on DataCite.

    ASSUMPTION: Record already exists on DataCite and should be updated.

    Function converts EnviDat record to DataCite Metadata Schema 4.4, for
    documentation: https://schema.datacite.org/meta/kernel-4.4/

    For documentation of DataCite API: https://support.datacite.org/docs/api

    For DataCite API reference:
    https://support.datacite.org/reference/introduction

    For documentation of DataCite API endpoint that updates DOIs:
    https://support.datacite.org/reference/put_dois-id

    Args:
        name (str): EnviDat records name that should be updated.
            Example: "mountland-jura"
        dc_dois (list[str]) :  List of DOIs with a specified DOI prefix in
          DataCite. This prefix is assigned in the config to the EnviDat
          prefix in DataCite.
          This arg is used during update of
          all records in datacite_update_all_records()
          Default value is None.

    Returns:
        dict/None: Returns dictionary with DataCite response data.
            If update fails then returns None.
    """
    try:
        # Get EnviDat record from CKAN API
        envidat_record = get_envidat_record(name)

        # Extract result from record
        result = envidat_record.get("result")

        # Check that result is truthy
        if not result:
            log.error(
                f"Failed to get 'result' from EnviDat record with " f"name '{name}'"
            )
            return

        # Check status code of response from call to CKAN API
        status_code = envidat_record.get("status_code")
        if status_code != 200:
            log.error(
                f"Failed to get EnviDat record with "
                f"name:  '{name}', "
                f"status_code:  {status_code}, "
                f"error:  {result}"
            )
            return

        # If dc_dois is truthy then check if doi not in dc_dois
        if dc_dois:
            doi = result.get("doi")
            if doi not in dc_dois:
                log.error(
                    f"EnviDat record with name '{name}' "
                    f"and doi '{doi} not in DataCite DOIs"
                )
                return

        # Update record with existing DOI in DataCite
        dc_response = publish_datacite(result, is_update=True)

        # Check that dc_response is dictionary
        if type(dc_response) != dict:
            log.error(
                f"Failed to update EnviDat record with name '"
                f"{name}', check console logs"
            )
            return

        # Add record name to dc_response
        dc_response["name"] = name

        # Return response for updated record, expecting a status code of 200
        if dc_response.get("status_code") == 200:
            return dc_response
        # Else log response for unexpected DataCite response status codes
        else:
            log.error(
                f"Failed to update record in DataCite with "
                f"name '{name}', error:  {dc_response}"
            )
            return

    except AttributeError as e:
        log.error(
            f"Failed to update record in DataCite with name '{name}', " f"error: {e}"
        )
        return


def get_dc_dois(num_records: int = 10000) -> list[str] | None:
    """Return a list of DOIs with a specified DOI prefix in DataCite.

    "DOI_PREFIX" in config is set to prefix assigned to EnviDat in DataCite.

    For DataCite API documentation of endpoint to get list of DOIs see:
    https://support.datacite.org/docs/api-get-lists

    Args:
        num_records (int): Number of records to retrieve from DORA API.
        Default value is 10000.
    """
    # Extract variables from config needed to call DataCite API
    # NOTE: List of DOIs should not be obtained from DataCite "test" API
    try:
        api_url = os.environ("DOI_API_URL")
        prefix = os.environ("DOI_PREFIX")
    except KeyError as e:
        log.error(f"KeyError: {e} does not exist in config")
        return None

    # Add prefix query param to url
    # Add page[size] param retrieve up to 10000 (default) records on the page
    api_url = f"{api_url}?prefix={prefix}&page[size]={num_records}"

    # Call API
    r = requests.get(api_url)

    # Return DOIs is successful
    if r.status_code == 200:

        # Get DOIs stored in record "id" values
        data = r.json().get("data")
        dois = [record.get("id") for record in data]

        # Return DOIs
        if dois:
            return dois
        else:
            log.error("Failed to get DOIs")
            return None

    # Else log error message and return none
    else:
        log.error("Failed to get DOIs")
        return None


# TODO implement error handling (try/excpet)
def get_published_record_names_with_dois() -> list[dict] | None:
    """Return EnviDat record names that have a DOI and a "publication_state"
    value of "published".

    Logs records that do not have a DOI or have a DOI but a
    "publication_state" value of "reserved".

    Returns: list[dict]/None: List of record name dictionaries with "name"
              and "doi" keys. Returns None if failed to obtain list.
    """
    err_message = "Failed to get names of published records with DOIs."

    # Get JSON response from call to CKAN
    # "API_CURRENT_PACKAGE_LIST_WITH_RESOURCES" endpoint
    response_json = get_response_json(
        api_host="API_HOST",
        api_path="API_CURRENT_PACKAGE_LIST_WITH_RESOURCES",
        query={"limit": 100000},
    )

    # Extract and return record names from records that have a DOI and are
    # published
    if response_json:

        records = response_json["result"]
        if records:
            published_records = []

            for record in records:
                # TODO review condition
                if record.get("doi") and record.get("publication_state") == "published":
                    published_records.append(
                        {"name": record.get("name"), "doi": record.get("doi")}
                    )

                # Log records that have a value for the "doi" key
                # and a "reserved" value for the "publication_state" value
                # but are not in dc_dois
                elif (
                    record.get("doi") and record.get("publication_state") == "reserved"
                ):
                    log.warning(
                        f"Record '{record.get('name')}' "
                        f"with DOI {record['doi']} "
                        f" has a 'publication_state' value of"
                        f" '{record['publication_state']}' and is not "
                        f"published on "
                        f"DataCite"
                    )

                # Log warning for records that have a "doi" value of empty
                # string ""
                elif record.get("doi") == "":
                    log.warning(f"Record does not have a DOI:'{record.get('name')}'")

            return published_records

        else:
            log.error(err_message)
            return None

    else:
        log.error(err_message)
        return None



"""Council information system (RIS) of the municipal council of Zurich."""

import json
from logging import getLogger

from envidat.api.v1 import get_protocol_and_domain

log = getLogger(__name__)


def convert_ris(metadata_record: dict) -> str:
    """Generate output string in RIS format.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.

    Returns:
        str: string in RIS format

    """
    try:
        return ris_convert_dataset(metadata_record)
    except ValueError as e:
        log.error(e)
        log.error("Cannot convert package to RIS format.")
        raise ValueError("Failed to convert package to RIS format.")


def ris_convert_dataset(dataset_dict: dict) -> str:
    """Create the RIS string from API dictionary."""
    ris_list = []

    #   TY  - DATA
    ris_list += ["TY  - DATA"]

    #   T1  - Title
    title = dataset_dict["title"]
    ris_list += ["T1  - " + title]

    #   AU  - Authors
    authors = json.loads(dataset_dict.get("author", "[]"))
    for author in authors:
        author_name = ""
        if author.get("given_name"):
            author_name += author["given_name"].strip() + " "
        author_name += author["name"].strip()
        ris_list += ["AU  - " + author_name]

    #   DO  - DOI
    doi = dataset_dict.get("doi", "").strip()
    if doi:
        ris_list += ["DO  - " + doi]

    #   UR  - dataset url as information
    protocol, host = get_protocol_and_domain()
    package_name = dataset_dict.get("name", "")
    package_url = f"{protocol}://{host}/dataset/{package_name}"
    ris_list += ["UR  - " + package_url]

    #   KW  - keywords
    keywords = get_keywords(dataset_dict)
    for keyword in keywords:
        ris_list += ["KW  - " + keyword]

    #   PY  - publication year
    publication = json.loads(dataset_dict.get("publication", "{}"))
    publication_year = publication["publication_year"]
    ris_list += ["PY  - " + publication_year]

    #   PB  - Publisher
    publisher = publication["publisher"]
    ris_list += ["PB  - " + publisher]

    #   LA  - en
    language = dataset_dict.get("language", "en").strip()
    if len(language) <= 0:
        language = "en"
    ris_list += ["LA  - " + language]

    #   ER  -
    ris_list += ["ER  - "]

    return "\n".join(ris_list)


def get_keywords(data_dict: dict):
    """Extract keywords from tags."""
    keywords = []
    for tag in data_dict.get("tags", []):
        name = tag.get("display_name", "").upper()
        keywords += [name]
    return keywords



"""BibTeX reference management format, for LaTeX formatting."""

import json
from logging import getLogger

from envidat.api.v1 import get_protocol_and_domain

log = getLogger(__name__)


def convert_bibtex(metadata_record: dict) -> str:
    """Generate output string in BibTex format.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.

    Returns:
        str: string in BibTeX format

    """
    try:
        return bibtex_convert_dataset(
            metadata_record,
        )
    except AttributeError as e:
        log.error(e)
        log.error("Cannot convert package to BibTeX format.")
        raise AttributeError("Failed to convert package to BibTeX format.")


def bibtex_convert_dataset(dataset: dict) -> str:
    """Create the BibTeX string from API dictionary."""
    # name as identifier (plus year later)
    name = dataset["name"]
    converted_package = "@misc { " + name

    # year (add to name) and journal
    publication = json.loads(dataset.get("publication", "{}"))
    publication_year = publication["publication_year"]

    converted_package += f"-{publication_year}"
    converted_package += f',\n\t year = "{publication_year}"'
    publisher = publication["publisher"]
    converted_package += f',\n\t publisher = "{publisher}"'

    # title
    title = dataset["title"]
    converted_package += ',\n\t title = "' + title + '"'

    # author
    authors = json.loads(dataset.get("author", "[]"))
    author_names = []
    for author in authors:
        author_name = ""
        if author.get("given_name"):
            author_name += author["given_name"].strip() + " "
        author_names += [author_name + author["name"]]
    bibtex_author = " and ".join(author_names)
    converted_package += f',\n\t author = "{bibtex_author}"'

    # DOI
    doi = dataset.get("doi", "").strip()
    if doi:
        converted_package += f',\n\t DOI = "http://dx.doi.org/{doi}"'

    # url
    protocol, host = get_protocol_and_domain()
    package_name = dataset.get("name", "")
    url = f"{protocol}://{host}/dataset/{package_name}"
    converted_package += ',\n\t url = "' + url + '"'

    # close bracket
    converted_package += "\n\t}"

    return converted_package



"""Standard XML format for NASA and other providers."""


import logging

from xmltodict import unparse

from envidat.api.v1 import get_metadata_json_with_resources

log = logging.getLogger(__name__)


def convert_xml(package: dict) -> str:
    """Convert EnviDat record to XML format.

    Args:
        package (dict): Package JSON from API.

    Returns:
        str: XML formatted string.

    """
    root = {"root": package}

    # Try to convert packages (i.e. records with metadata) to XML format
    try:
        converted_data_xml = unparse(root, short_empty_elements=True, pretty=True)
        return converted_data_xml

    except Exception as e:
        log.error("ERROR: Cannot convert EnviDat packages to XML format.")
        log.error(e)


def convert_xml_all_resources() -> str:
    """Convert EnviDat JSON records to XML format.

    Returns:
        str: XML formatted string.

    Note:
        Only valid for metadata schema of EnviDat.
    """
    metadata_json = get_metadata_json_with_resources()

    # TODO verify which root element is required for XML format
    metadata_json = {"root": metadata_json}

    # Try to convert packages (i.e. records with metadata) to XML format
    try:
        converted_data_xml = unparse(
            metadata_json, short_empty_elements=True, pretty=True
        )
        return converted_data_xml

    except Exception as e:
        log.error("ERROR: Cannot convert EnviDat packages to XML format.")
        log.error(e)
